/*
 *  hemca
 *    --- device driver for Hoshin HE1442 Multi-Channel Analyzer
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/irq.h>
#ifdef MODULE
#include <linux/module.h>
#define HEMCA_INIT_RET(x)  return x
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#define HEMCA_INIT_RET(x)  return kmem_start
#endif

#include <linux/version.h>
#if   LINUX_VERSION_CODE >= 0x020100
#  include <asm/uaccess.h>
#  define LINUX_2_2
#  define relse_t int
#  define HEMCA_MINOR     MINOR(file->f_dentry->d_inode->i_rdev)
#elif LINUX_VERSION_CODE >= 0x020000
#  include <asm/segment.h>
#  define LINUX_2_0
#  define relse_t void
#  define HEMCA_MINOR     MINOR(inode->i_rdev)
#  define copy_from_user  memcpy_fromfs
#  define copy_to_user    memcpy_tofs
#  define signal_pending(current)  (current->signal & ~current->blocked)
#else
#  error "This kernel is too old: not supported by this file"
#endif

#include "hemca.h"
#define REG (HEMCA_PORT+0)
#define DAT (HEMCA_PORT+1)

static unsigned short hemca_buff[HEMCA_MAXBUF]; /* ADC data buffer */
static unsigned int   hemca_rp=0, hemca_wp=0;   /* read & write pointer */

static struct wait_queue *wait_que=NULL;
static int    wakeups=0;
static int    hemca_intr_flag=0;
static long   hemca_lost=0;

static unsigned int hemca_hist[HEMCA_CHAN];	/* histogram buffer */

static int     hemca_open    (struct inode *inode, struct file *file) {
  MOD_INC_USE_COUNT;
  return 0;
}

static relse_t hemca_release (struct inode *inode, struct file *file) {
  MOD_DEC_USE_COUNT;
#ifdef LINUX_2_2
  return 0;
#endif
}

#if   defined(LINUX_2_2)
static ssize_t hemca_read    (struct file *file, char *buffer, 
				size_t count, loff_t *fpos) {
#elif defined(LINUX_2_0)
static int     hemca_read    (struct inode *inode, struct file *file, 
				char *buffer, int count) {
#endif
  int nbytes;
  if (HEMCA_MINOR==0) {			/* minor=0: histogram in binary */
    nbytes=HEMCA_CHAN*sizeof(int);
    if (nbytes>count) nbytes=(count/sizeof(int))*sizeof(int);
    copy_to_user((void *)buffer,&hemca_hist[0],nbytes);
  } else {				/* minor=1: raw-data in binary */
    int real_rp, remain;
    if (hemca_wp==hemca_rp) {
      wakeups++;
      interruptible_sleep_on(&wait_que);
      if (wakeups) wakeups--;
      if (signal_pending(current)) return -EINTR;
    }
    nbytes=(hemca_wp-hemca_rp)*sizeof(short);	/* bytes of available data */
    if (nbytes>count) nbytes=(count/sizeof(short))*sizeof(short);
    real_rp=hemca_rp%HEMCA_MAXBUF;		/* real buffer pointer */
    if (nbytes>(remain=(HEMCA_MAXBUF-real_rp)*sizeof(short))) {
      copy_to_user((void *)buffer,&hemca_buff[real_rp],remain);
      copy_to_user((void *)buffer,&hemca_buff[real_rp],nbytes-remain);
    } else {
      copy_to_user((void *)buffer,&hemca_buff[real_rp],nbytes);
    }
    hemca_rp+=nbytes/sizeof(short);
  }
  return nbytes;
}

static void   hemca_intr     (int irq, void *dev_id, struct pt_regs *regs) {
  unsigned char sta;
  unsigned int  dat;
  if ((hemca_wp-hemca_rp)>=HEMCA_MAXBUF) {hemca_lost++;} /* overwrite & lost */
  outb(1,REG);	/* register = lower 4-bit data & status */
  sta=inb(DAT);	/* read lower 4-bit data & status */
  if ((sta&0x80)==0) return;	/* if not data ready */
  outb(0,REG);	/* register = upper 8-bit data */
  /* read upper 8-bit data and merge to lower 4-bit */
  dat=((unsigned short)inb(DAT)<<4)+(unsigned short)(sta&0x0F);
  /* this also clears busy status and makes ADC ready for next event */
  hemca_buff[(hemca_wp++)%HEMCA_MAXBUF]=dat;  /* store list data */
  hemca_hist[(dat%4096)/(4096/HEMCA_CHAN)]++; /* count up histogram */
  if ((hemca_wp!=hemca_rp) && (wakeups!=0)) wake_up(&wait_que);
};

static int   hemca_ioctl     (struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg) {
  int i,ans=0;
  unsigned char sta;
  unsigned long time, flags;
  switch (cmd) {
  case HEMCA_START:	/* start data aquisition (enable interrupt), no arg */
    hemca_lost=0;	/* clear lost counter */
    save_flags(flags);
    cli();
    outb(2,REG);	/* register = lower 8-bit timer */
    outb(0,DAT);	/* timer    = 0 */
    outb(0,REG);	/* register = upper 8-bit data */
    sta=inb(DAT);	/* dummy read to clear busy status */
    ans=request_irq(HEMCA_IRQ,hemca_intr,SA_INTERRUPT,HEMCA_HARDWARE,NULL);
    if (ans==0) hemca_intr_flag=1;	/* enable interrupt */
    restore_flags(flags);
    return 0;
  case HEMCA_STOP:	/* stop data acquisition (disable interrupt), no arg */
    cli();
    if (hemca_intr_flag!=0) free_irq(HEMCA_IRQ,NULL);
    hemca_intr_flag=0;	/* disable interrupt */
    sti();
    if (wakeups!=0) wake_up_interruptible(&wait_que);
    return 0;
  case HEMCA_CLEAR:	/* clear histogram, no arg */
    for (i=0;i<HEMCA_CHAN;i++) hemca_hist[i]=0;
    return 0;
  case HEMCA_TIME:	/* read live time and return as long */
    save_flags(flags);
    cli();
    outb(2,REG);	/* register = lower 8-bit timer */
    sta=inb(DAT);	/* read lower 8-bit time */
    outb(3,REG);	/* register = upper 8-bit time */
    time=((unsigned long)inb(DAT)<<8)+(unsigned long)sta;
    /* read upper 8-bit time and merge, giving true time */
    restore_flags(flags);
    put_user(time,(long*)arg);
    return 0;
  case HEMCA_LOST:	/* return lost count as log */
    put_user(hemca_lost,(long*)arg);
    return 0;
  default:
    printk("hemca: invalid ioctl (%d)\n",cmd);
    return -ENOIOCTLCMD;
  }
}

static struct file_operations hemca_fops = {
  NULL	/* hemca_lseek */,
  hemca_read,
  NULL	/* hemca_write */,
  NULL	/* hemca_readdir */,
  NULL	/* hemca_select */,
  hemca_ioctl,
  NULL	/* hemca_mmap */,
  hemca_open,
#ifdef LINUX_2_2
  NULL	/* hemca_flush */,
#endif
  hemca_release
};	/* if not defined, replace to NULL. */

#ifndef MODULE
unsigned long hemca_init(unsigned long kmem_start) {
#else
int init_module(void) {
#endif
  printk(HEMCA_DEVICE ": device installed.\n");
  printk(HEMCA_DEVICE ": IO base = 0x%X  IRQ = %d\n", HEMCA_PORT, HEMCA_IRQ);
  printk(HEMCA_DEVICE ": histogram size is %d channel\n", HEMCA_CHAN);
  printk(HEMCA_DEVICE ": list data buffer is %d bytes\n",
	 HEMCA_MAXBUF*sizeof(short));
  if (register_chrdev(HEMCA_NO,HEMCA_DEVICE,&hemca_fops)) {
    printk(HEMCA_DEVICE
	   ": error -- cannot register to major device %d !\n",HEMCA_NO);
    HEMCA_INIT_RET(-ENODEV);
  }
  if (check_region(HEMCA_PORT,2)) {
    printk(HEMCA_DEVICE
	   ": error -- cannot register I/O port %x !\n",HEMCA_PORT);
    unregister_chrdev(HEMCA_NO,HEMCA_DEVICE);
    HEMCA_INIT_RET(-ENODEV);
  }
  request_region(HEMCA_PORT,2,HEMCA_DEVICE);
  if (request_irq(HEMCA_IRQ,hemca_intr,SA_INTERRUPT,HEMCA_HARDWARE,NULL)) {
    printk(HEMCA_DEVICE
	   ": error -- cannot register IRQ %d !\n",HEMCA_IRQ);
    release_region(HEMCA_PORT,2);
    unregister_chrdev(HEMCA_NO,HEMCA_DEVICE);
    HEMCA_INIT_RET(-ENODEV);
  }
  free_irq(HEMCA_IRQ,NULL);
  HEMCA_INIT_RET(0);
}
#ifdef  MODULE
void cleanup_module(void) {
  printk(HEMCA_DEVICE ": device removed.\n");
  if (hemca_intr_flag!=0) free_irq(HEMCA_IRQ,NULL);
  release_region(HEMCA_PORT,2);
  unregister_chrdev(HEMCA_NO,HEMCA_DEVICE);
}
#endif
